'use strict';

var _couchbase = require('couchbase');

var _couchbase2 = _interopRequireDefault(_couchbase);

var _bluebird = require('bluebird');

var _bluebird2 = _interopRequireDefault(_bluebird);

var _dbMigrateBase = require('db-migrate-base');

var _dbMigrateBase2 = _interopRequireDefault(_dbMigrateBase);

var _bunyan = require('bunyan');

var _bunyan2 = _interopRequireDefault(_bunyan);

var _bunyanPrettystream = require('bunyan-prettystream');

var _bunyanPrettystream2 = _interopRequireDefault(_bunyanPrettystream);

var _moment = require('moment');

var _moment2 = _interopRequireDefault(_moment);

var _deasync = require('deasync');

var _deasync2 = _interopRequireDefault(_deasync);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var ottoman = require('ottoman'); // Directions:
// https://db-migrate.readthedocs.io/en/latest/Developers/contributing/#creating-your-own-driver
//
// Couchbase driver docs:
// http://docs.couchbase.com/sdk-api/couchbase-node-client-2.1.4/Cluster.html
//
// Conventions:
// driver's noSQL API refers to "collections" which Couchbase doesn't have.  We
// use buckets to be collections.
// RDBMS "Tables" are Ottoman models.
// RDBMS "Rows" are ottoman model instances (documents)
// RDBMS "Columns" are paths into an ottoman model.  Not just attributes, but paths.
// RDBMS "Databases" are Couchbase buckets.


var prettyStdOut = new _bunyanPrettystream2.default();
prettyStdOut.pipe(process.stdout);

/**
 * contains the standard datatypes of db-migrate
 */
var type = null;

var UNSUPPORTED = 'operation not supported';
var ottomanType = '_type';

var log = _bunyan2.default.createLogger({
  name: 'db-migrate-couchbase',
  streams: [{
    level: 'debug',
    type: 'raw',
    stream: prettyStdOut
  }]
});

var internalLogger = null;

var DEFAULTS = {
  host: process.env.COUCHBASE_HOST || 'couchbase://localhost',
  port: 8091,
  user: 'admin',
  password: 'admin001*'
};

// Escape n1ql identifiers so they are syntactically valid even if they
// contain reserved words.  So foo.bar becomes `foo`.`bar`
var n1qlEscape = function n1qlEscape(param) {
  return param.split('.').map(function (p) {
    return '`' + p + '`';
  }).join('.');
};

var singleton = null;

var CouchbaseDriver = _dbMigrateBase2.default.extend({
  init: function init(connection, internals, config) {
    var _this = this;

    if (singleton) {
      log.warning('You should only create one instance of the db-migrate-couchbase driver');
      return singleton;
    }

    if (internals && internals.argv && internals.argv['sql-file']) {
      throw new Error('This driver does not support the --sql-file option.');
    }

    if (internals && internals.migrationTable !== 'migrations') {
      log.warn('Driver ignores migration table option; it uses ottoman model MigrationRun');
      internals.migrationTable = 'migrations';
    }

    if (internals && internals.seedTable !== 'seeds') {
      log.warn('Driver ignores seed table option; it uses ottoman model MigrationSeed');
      internals.seedTable = 'seeds';
    }

    log.debug('init');
    // console.log(internals);
    this._super(internals);
    this.connection = connection;
    this.user = config.user;
    this.password = config.password;
    this.bucketPassword = config.bucketPassword;
    this.manager = connection.manager(this.user, this.password);
    this.models = {};

    if (!config.migrationBucket) {
      throw new Error('Configuration must specify migrationBucket');
    }

    var errCallback = function errCallback(err) {
      if (err) {
        throw new Error('Could not open migration bucket: ' + err);
      }
    };

    this.migrationBucket = this.bucketPassword ? this.connection.openBucket(config.migrationBucket, this.bucketPassword, errCallback) : this.connection.openBucket(config.migrationBucket, errCallback);

    this.active = this.migrationBucket;
    ottoman.bucket = this.migrationBucket;
    this.migrationAdapter = new ottoman.StoreAdapter.Couchbase(this.migrationBucket);

    // Property names are obligatory here; the migrate framework expects exactly these,
    // "name" and "run_on"
    this.MigrationRun = ottoman.model('MigrationRun', {
      oid: { type: 'string', auto: 'uuid', readonly: true },
      name: { type: 'string' },
      run_on: { type: 'Date' }
    }, {
      id: 'oid'
      // store: this.migrationAdapter,
    });

    this.MigrationSeed = ottoman.model('MigrationSeed', {
      oid: { type: 'string', auto: 'uuid', readonly: true },
      name: { type: 'string' },
      run_on: { type: 'Date' },
      ifNotExists: { type: 'boolean', default: true }
    }, {
      id: 'oid'
    });

    this.ready = false;

    ottoman.ensureIndices(function (err) {
      if (err) {
        log.error('Failed to create ottoman indices', { err: err });
      }

      _this.ready = true;
    });

    this.models.MigrationRun = this.MigrationRun;

    singleton = this;

    return singleton;
  },

  activeBucketName: function activeBucketName() {
    return singleton.active._name;
  },

  runN1ql: function runN1ql(query, params, callback) {
    var n1ql = _couchbase2.default.N1qlQuery.fromString(query);

    // Use strong consistency on all migration queries, to avoid errors or missing records
    // owing from cluster synchronization.
    n1ql.consistency(_couchbase2.default.N1qlQuery.Consistency.STATEMENT_PLUS);

    if (!singleton.active) {
      return _bluebird2.default.reject('no active bucket').nodeify(callback);
    }

    if (singleton.internals.dryRun) {
      log.info('DRY RUN', { query: query, params: params });
      return _bluebird2.default.resolve(true).nodeify(callback);
    }

    return new _bluebird2.default(function (resolve, reject) {
      log.info(query);
      return singleton.active.query(n1ql, params, function (err, rows, meta) {
        if (err) {
          log.error('n1ql error', { err: err });
          return reject(err);
        }

        return resolve({ rows: rows, meta: meta });
      });
    }).nodeify(callback);
  },

  close: function close(callback) {
    log.info('close');

    // Couchbase doesn't disconnect nicely, leaves open threads.
    setTimeout(function () {
      log.info('Time!');
      process.exit(0);
    }, 5000);

    callback();
  },

  /**
   * Provided for compatibility with the driver, but this is an alias for runN1ql
   */
  runSql: function runSql(query, params, callback) {
    log.info('runSql', { query: query, params: params });
    return singleton.runN1ql(query, params, callback);
  },

  /**
   * Provided for compatibility with the driver, but this is an alias for runN1ql
   */
  all: function all(query, params, callback) {
    log.info('all', { query: query, params: params });
    return singleton.runN1ql(query, params, callback);
  },

  /**
   * Queries the migrations table
   *
   * @param callback
   */
  allLoadedMigrations: function allLoadedMigrations(callback) {
    log.info('allLoadedMigrations');

    return new _bluebird2.default(function (resolve, reject) {
      singleton.getModel('MigrationRun').find({}, {
        sort: ['run_on DESC', 'name DESC'],
        consistency: ottoman.Consistency.GLOBAL
      }, function (err, models) {
        if (err) {
          return reject(err);
        }
        return resolve(models);
      });
    }).nodeify(callback);
  },

  /**
   * Deletes a migration
   *
   * @param migrationName   - The name of the migration to be deleted
   * @param callback
   */
  deleteMigration: function deleteMigration(migrationName, callback) {
    log.info('deleteMigration', { migrationName: migrationName });

    if (singleton.internals.dryRun) {
      return _bluebird2.default.resolve(true).nodeify(callback);
    }

    return new _bluebird2.default(function (resolve, reject) {
      singleton.MigrationRun.find({
        name: migrationName
      }, {}, function (err, rows) {
        if (err) {
          return reject(err);
        }

        if (rows.length === 0) {
          return resolve(false);
        }

        var deletePromises = rows.map(function (migRecord) {
          return new _bluebird2.default(function (delResolve, delReject) {
            log.info('Deleting migration record ', { migRecord: migRecord });
            return migRecord.remove(function (removeErr) {
              if (removeErr) {
                return delReject(removeErr);
              }
              return delResolve(true);
            });
          });
        });

        return _bluebird2.default.all(deletePromises).then(function (r) {
          return resolve(r);
        });
      });
    }).nodeify(callback);
  },

  mapDataType: function mapDataType(spec) {
    var map = {};

    var ottomanTypes = ['string', 'integer', 'Date', 'number', 'boolean', 'Mixed'];

    // Already an ottoman type?
    if (ottomanTypes.indexOf(spec.type) !== -1) {
      return spec.type;
    }

    map[type.TEXT] = map[type.CHAR] = map[type.STRING] = 'string';
    map[type.INTEGER] = map[type.SMALLINT] = 'integer';
    map[type.BIGINT] = 'string';
    map[type.REAL] = map[type.DECIMAL] = map[type.REAL] = 'number';
    map[type.BOOLEAN] = 'boolean';
    map[type.DATE] = map[type.DATE_TIME] = 'Date';
    map[type.BLOB] = 'string';
    map[type.TIME] = 'string';
    map[type.BINARY] = 'string';

    if (spec.type in map) {
      return map[spec.type];
    }

    return 'string';
  },

  /**
   * Unsupported; if Couchbase supported transactions, here's where
   * START TRANSACTION would go.
   */
  startMigration: function startMigration(cb) {
    log.info('startMigration');
    return _bluebird2.default.resolve(true).nodeify(cb);
  },

  /**
   * Unsupported; if Couchbase supported transactions, here's where
   * COMMIT would go.
   */
  endMigration: function endMigration(cb) {
    log.info('endMigration');
    return _bluebird2.default.resolve(true).nodeify(cb);
  },

  /**
   * Remembers a model for later reference.
   * @param modelName name of the model
   * @param ottomanModel the actual model
   * @returns the driver instance for chaining.
   */
  registerModel: function registerModel(modelName, ottomanModel) {
    singleton.models[modelName] = ottomanModel;
    return singleton;
  },

  /**
   * Return a model that this migration driver knows about
   * @param modelName the name of the model.
   */
  getModel: function getModel(modelName) {
    return singleton.models[modelName];
  },

  /**
   * Create table: just an alias for making an ottoman model
   */
  createTable: function createTable(ottomanModelName, schema, callback) {
    if (ottomanModelName === 'migration') {
      return _bluebird2.default.resolve(singleton.MigrationRun).nodeify(callback);
    }

    if (ottomanModelName === 'seeds') {
      return _bluebird2.default.resolve(singleton.MigrationSeed).nodeify(callback);
    }

    log.info('create table / ottoman model', { ottomanModelName: ottomanModelName, schema: schema });

    // This is a special case, we store migrations under ottoman model MigrationRun.
    // Also, createTable when migration starts creates a table with a field called 'id',
    // which isn't permitted in ottoman.
    if (ottomanModelName === 'migrations') {
      return _bluebird2.default.resolve(this.MigrationRun).nodeify(callback);
    }

    var model = ottoman.model(ottomanModelName, schema);
    singleton.registerModel(ottomanModelName, model);

    return _bluebird2.default.resolve(singleton.getModel(ottomanModelName)).nodeify(callback);
  },

  /**
   * Create an ottoman model; alias for createTable
   */
  createOttomanModel: function createOttomanModel(ottomanModelName, options, callback) {
    return singleton.createTable(ottomanModelName, options, callback);
  },

  /** unsupported */
  dropTable: function dropTable(tableName, options, callback) {
    log.info('dropTable', { tableName: tableName, options: options });
    return _bluebird2.default.reject(UNSUPPORTED).nodeify(callback);
  },

  dropOttomanModel: function dropOttomanModel(ottomanModelName, options, callback) {
    return singleton.dropTable(ottomanModelName, options, callback);
  },

  /**
   * Unsupported; renaming a model is easy, but carrying with it and pruning its indexes is hard.
   */
  renameTable: function renameTable(ottomanModelName, newOttomanModelName, callback) {
    log.info('renameTable', { ottomanModelName: ottomanModelName, newOttomanModelName: newOttomanModelName });
    return _bluebird2.default.reject(UNSUPPORTED).nodeify(callback);
  },

  renameOttomanModel: function renameOttomanModel(ottomanModelName, newOttomanModelName, callback) {
    return singleton.renameTable(ottomanModelName, newOttomanModelName, callback);
  },

  addColumn: function addColumn(ottomanModelName, modelPath, pathSpec, callback) {
    log.info('addColumn', { ottomanModelName: ottomanModelName, modelPath: modelPath, pathSpec: pathSpec });

    if (!ottomanModelName) {
      return _bluebird2.default.reject('missing ottoman model name').nodeify(callback);
    }

    if (!modelPath) {
      return _bluebird2.default.reject('missing model path').nodeify(callback);
    }

    // TODO -- default value of null won't work for all ottoman types, like
    // Mixed, ref, etc.   But it will work for almost all primitive types.
    // const ottomanType = this.mapDataType(pathSpec);

    var q = '\n      UPDATE `' + singleton.active._name + '`\n      SET ' + n1qlEscape(modelPath) + ' = null\n      WHERE `' + ottomanType + '` = \'' + ottomanModelName + '\'';

    return singleton.runN1ql(q, {}, callback);
  },

  addOttomanPath: function addOttomanPath(ottomanModelName, modelPath, pathSpec, callback) {
    return singleton.addColumn(ottomanModelName, modelPath, pathSpec, callback);
  },

  removeColumn: function removeColumn(ottomanModelName, modelPath, callback) {
    log.info('removeColumn', { ottomanModelName: ottomanModelName, modelPath: modelPath });

    var q = '\n      UPDATE `' + singleton.active._name + '`\n      UNSET ' + n1qlEscape(modelPath) + '\n      WHERE `' + ottomanType + '` = \'' + ottomanModelName + '\'';

    return singleton.runN1ql(q, {}, callback);
  },

  removeOttomanPath: function removeOttomanPath(ottomanModelName, modelPath, callback) {
    return singleton.removeColumn(ottomanModelName, modelPath, callback);
  },

  /**
   * alias
   * @see renameOttomanPath
   */
  renameColumn: function renameColumn(ottomanModelName, oldModelPath, newModelPath, callback) {
    log.info('renameColumn', { ottomanModelName: ottomanModelName, oldModelPath: oldModelPath, newModelPath: newModelPath });

    var q = '\n      UPDATE `' + singleton.active._name + '`\n      SET ' + n1qlEscape(newModelPath) + ' = ' + n1qlEscape(oldModelPath) + '\n      UNSET ' + n1qlEscape(oldModelPath) + '\n      WHERE `' + ottomanType + '` = \'' + ottomanModelName + '\'';

    log.info('renameColumn', { q: q });
    return singleton.runN1ql(q, {}, callback);
  },

  /**
   * Use this function with care, it does **not** adjust any indexes that might already be
   * on the path, so consider dropping those before doing this.
   */
  renameOttomanPath: function renameOttomanPath(ottomanModelName, oldModelPath, newModelPath, callback) {
    return singleton.renameColumn(ottomanModelName, oldModelPath, newModelPath, callback);
  },

  changeColumn: function changeColumn(ottomanModelName, ottomanPathName, pathSpec, callback) {
    log.info('changeColumn', { ottomanModelName: ottomanModelName, ottomanPathName: ottomanPathName, pathSpec: pathSpec });
    return _bluebird2.default.reject(UNSUPPORTED).nodeify(callback);
  },

  /**
   * Couchbase doesn't support this as such, because the concepts of columnNameArrays
   * and valueArrays aren't really a good fit.
   */
  insert: function insert(tableName, columnNameArray, valueArray, callback) {
    log.info('insert', { tableName: tableName, columnNameArray: columnNameArray, valueArray: valueArray });
    return _bluebird2.default.reject(UNSUPPORTED).nodeify(callback);
  },

  addMigrationRecord: function addMigrationRecord(name, callback) {
    log.info('addMigrationRecord', { name: name });
    var i = new singleton.MigrationRun({
      name: name,
      run_on: (0, _moment2.default)().utc()
    });

    if (singleton.internals.dryRun) {
      return _bluebird2.default.resolve(true).nodeify(callback);
    }

    return new _bluebird2.default(function (resolve, reject) {
      return i.save(function (err) {
        if (err) {
          return reject(err);
        }
        return resolve(true);
      });
    }).nodeify(callback);
  },

  /**
   * Creates a bucket with a given name, and sets it to be active.
   */
  createBucket: function createBucket(bucketName, options, callback) {
    log.info('createBucket', { bucketName: bucketName });

    return new _bluebird2.default(function (resolve, reject) {
      return singleton.manager.createBucket(bucketName, options || {}, function (err) {
        if (err) {
          return reject(err);
        }
        log.info('Created bucket', { bucketName: bucketName, err: err });
        singleton.active = singleton.connection.openBucket(bucketName);
        return resolve(singleton.active);
      });
    }).nodeify(callback);
  },

  /**
   * Drops a bucket.   Note, due to couchbase server implementation and what it
   * takes to drop a bucket, this may take quite some time to resolve (> 1 min, even in
   * small cases).   Do not use this unless you know what you're doing, as this function
   * can destroy quite a lot very quickly.
   */
  dropBucket: function dropBucket(bucketName, callback) {
    log.info('dropBucket', { bucketName: bucketName });

    if (singleton.internals.dryRun) {
      return _bluebird2.default.resolve(true).nodeify(callback);
    }

    return new _bluebird2.default(function (resolve, reject) {
      return singleton.manager.removeBucket(bucketName, function (err) {
        if (err) {
          return reject(err);
        }

        // Remove active if we just remove that one, because it's a reference that's
        // no longer valid.
        if (singleton.active._name === bucketName) {
          singleton.active = null;
        }

        return resolve(true);
      });
    }).nodeify(callback);
  },

  changeType: function changeType(ottomanModelName, ottomanPath, oldType, newType, callback) {
    // For various conversions, store functions that perform the conversion on an
    // identifier.  This gets spliced into n1ql.
    // Strings rather than functions are error messages back to the user about why that
    // conversion may not make sense.
    var supported = {
      Date: {
        string: 'Dates and strings are already interchangable',
        number: 'not supported'
      },

      boolean: {
        string: function string(i) {
          return 'TOSTRING(' + i + ')';
        },
        number: 'Converting booleans to numbers doesn\'t really make sense'
      },

      string: {
        Date: 'Dates and strings are already interchangeable',
        boolean: function boolean(i) {
          return 'TOBOOLEAN(' + i + ')';
        },
        number: function number(i) {
          return 'TONUMBER(' + i + ')';
        }
      },

      number: {
        string: function string(i) {
          return 'TOSTRING(' + i + ')';
        },
        boolean: function boolean(i) {
          return 'TOBOOLEAN(' + i + ')';
        },
        Date: 'not supported'
      }
    };

    log.info('changeType', { ottomanModelName: ottomanModelName, ottomanPath: ottomanPath, oldType: oldType, newType: newType });

    // Error conditions to check, before we issue a query that could damage data.
    if (!ottomanModelName || !ottomanPath || !oldType || !newType) {
      return _bluebird2.default.reject('Missing required argument').nodeify(callback);
    } else if (oldType === newType) {
      return _bluebird2.default.reject('Cannot convert from one type to itself').nodeify(callback);
    } else if (!supported[oldType]) {
      return _bluebird2.default.reject('Conversion of type ' + oldType + ' is not supported');
    } else if (!supported[oldType][newType]) {
      return _bluebird2.default.reject('Type ' + oldType + ' does not have a supported conversion to ' + newType);
    }

    var conversion = supported[oldType][newType];
    if (typeof conversion === 'string') {
      return _bluebird2.default.reject('Can\'t convert ' + oldType + ' to ' + newType + ': ' + conversion).nodeify(callback);
    }

    var p = n1qlEscape(ottomanPath);
    var query = 'UPDATE `' + singleton.active._name + '`\n    SET ' + p + ' = ' + conversion(p) + '\n    WHERE `' + ottomanType + '`=\'' + ottomanModelName + '\'';

    return singleton.runN1ql(query);
  },

  /**
   * Adds an index to a collection
   * @param ottomanModelName the name of the ottoman model
   * @param indexName - The name of the index to add
   * @param columns - The columns to add an index on
   * @param callback
   */
  addIndex: function addIndex(ottomanModelName, indexName, columns, callback) {
    log.info('Add Index', { indexName: indexName, columns: columns, ottomanModelName: ottomanModelName });

    var srcCols = columns;

    if (typeof columns === 'string') {
      srcCols = [columns];
    }

    var cols = srcCols.map(function (c) {
      return n1qlEscape(c);
    }).join(', ');

    var n1ql = '\n      CREATE INDEX `' + indexName + '` ON `' + singleton.active._name + '` (\n        ' + cols + '\n      )\n      WHERE `' + ottomanType + '`=\'' + ottomanModelName + '\'';

    return singleton.runN1ql(n1ql, {}, callback);
  },

  /**
   * Removes an index from a collection
   *
   * @param collectionName  - The collection to remove the index
   * @param indexName       - The name of the index to remove
   * @param columns
   */
  removeIndex: function removeIndex(indexName, callback) {
    log.info('removeIndex', { indexName: indexName });
    return singleton.runN1ql('DROP INDEX `' + singleton.active._name + '`.`' + indexName + '`', {}, callback);
  },

  /**
   * Opens the specified bucket, sets it active, and returns it.
   * @param bucketName the name of the bucket to use
   * @returns self, for chaining
   */
  withBucket: function withBucket(bucketName) {
    var bucket = null;

    return new _bluebird2.default(function (resolve, reject) {
      var onBucketOpen = function onBucketOpen(err) {
        if (err) {
          log.error('Failed to open bucket', { bucketName: bucketName, err: err });
          return reject(err);
        }

        singleton.active = bucket;
        return resolve(singleton);
      };

      log.info('Opening active bucket', { bucketName: bucketName });

      bucket = singleton.bucketPassword ? singleton.connection.openBucket(bucketName, singleton.bucketPassword, onBucketOpen) : singleton.connection.openBucket(bucketName, onBucketOpen);
    });
  },

  getBucketNames: function getBucketNames(callback) {
    log.info('getBucketNames');

    return new _bluebird2.default(function (resolve, reject) {
      return singleton.manager.listBuckets(function (err, buckets) {
        if (err) {
          return reject(err);
        }

        // log.info('Buckets', { buckets });
        return resolve(buckets);
      });
    }).nodeify(callback);
  },

  /**
   * Gets all the indexes for the active bucket.
   *
   * @param callback
   */
  getIndexes: function getIndexes(callback) {
    log.info('getIndexes');
    return singleton.runN1ql('SELECT * FROM system:indexes', {}, callback);
  }
});

_bluebird2.default.promisifyAll(CouchbaseDriver);

/**
 * Gets a connection to couchbase
 *
 * @param config    - The config to connect to couchbase
 * @param callback  - The callback to call with a CouchbaseDriver object
 */
var connect = function connect(config, intern, callback) {
  console.log('CONNECT');
  internalLogger = intern.mod ? intern.mod.log || log : log;
  type = intern.mod.type;

  // log.info('Connect', { config, intern, type });

  var port = void 0;
  if (config.port === undefined) {
    port = DEFAULTS.port;
  } else {
    port = config.port;
  }

  /* Export functions to the interface that are not normally part
   * of the migration interface.
   */
  var exportable = ['runN1ql', 'registerModel', 'getModel', 'createOttomanModel', 'renameOttomanModel', 'addOttomanPath', 'removeOttomanPath', 'renameOttomanPath', 'createBucket', 'dropBucket', 'withBucket', 'getBucketNames', 'getIndexes'];

  exportable.forEach(function (f) {
    /* eslint no-param-reassign: "off" */
    if (intern.interfaces && intern.interfaces.MigratorInterface) {
      intern.interfaces.MigratorInterface[f] = function () {
        var _ref;

        return _bluebird2.default.reject('Not implemented').nodeify((_ref = arguments.length - 1, arguments.length <= _ref ? undefined : arguments[_ref]));
      };
    }
    return f;
  });

  var cluster = null;

  if (config.host === undefined) {
    cluster = DEFAULTS.host + ':' + port + '?detailed_errcodes=1';
  } else {
    cluster = config.host + ':' + port + '?detailed_errcodes=1';
  }

  log.info('Connecting to cluster', { cluster: cluster });
  var db = new _couchbase2.default.Cluster(cluster);

  var driver = new CouchbaseDriver(db, intern, config);

  _deasync2.default.loopWhile(function () {
    return !driver.ready;
  });

  if (config.bucket) {
    return driver.withBucket(config.bucket).then(function (d) {
      return callback(null, d);
    });
  }

  return callback(null, new CouchbaseDriver(db, intern, config));
};

module.exports = { connect: connect };
//# sourceMappingURL=maps/index.js.map
