'use strict';

var _chai = require('chai');

var _chai2 = _interopRequireDefault(_chai);

var _index = require('../index.js');

var _index2 = _interopRequireDefault(_index);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var expect = _chai2.default.expect;
var DBMigrate = require('db-migrate');

var driver = null;
var config = require('../database.json').dev;
var inst = DBMigrate.getInstance(true);

describe('db-migrate-couchbase', function () {
  this.timeout(60000);

  before(function (done) {
    var config = inst.config.dev;
    var intern = {
      mod: {
        type: inst.dataTypes
      },
      interfaces: {
        MigratorInterface: {}
      }
    };

    _index2.default.connect(config, intern, function (err, d) {
      if (err) {
        return done(err);
      }
      driver = d;
      return done();
    });
  });

  describe('Migrations', function () {
    before(function (done) {
      driver.addMigrationRecord('TestMigration').then(driver.startMigration).then(driver.allLoadedMigrations).then(function (migrations) {
        var mine = migrations.filter(function (m) {
          return m.name === 'TestMigration';
        });
        expect(mine).to.be.ok;
        expect(mine.length).to.be.above(0);
      }).then(function () {
        return done();
      }).catch(function (err) {
        return done(err);
      });
    });

    describe('Unsupported operations', function () {
      var shouldBeUnsupported = function shouldBeUnsupported(promise, done) {
        return promise.then(function (r) {
          return done('Unsupported operation succeeded unexpectedly: ' + r);
        }).catch(function (err) {
          expect(err).to.equal('operation not supported');
          done();
        });
      };

      it('should not support drop table', function (done) {
        shouldBeUnsupported(driver.dropTable('Whatever'), done);
      });

      it('should not support dropOttomanModel', function (done) {
        shouldBeUnsupported(driver.dropOttomanModel('Whatever'), done);
      });

      it('should not support renaming models', function (done) {
        shouldBeUnsupported(driver.renameOttomanModel('TestModel', 'OtherModel'), done);
      });

      it('should not support inserting', function (done) {
        shouldBeUnsupported(driver.insert('Whatever', ['a'], [1]), done);
      });

      it('should not support changing columns', function (done) {
        shouldBeUnsupported(driver.changeColumn('Whatever', 'orig', {}), done);
      });

      it('should not support renaming tables', function (done) {
        shouldBeUnsupported(driver.renameTable('TestModel', 'OtherModel'), done);
      });
    });

    describe('Models', function () {
      it('should create a model', function (done) {
        driver.createOttomanModel('Foo', {
          name: 'string'
        }, {}).then(function (model) {
          expect(model).to.be.ok;
          expect(model.schema).to.be.ok;
          done();
        }).catch(function (err) {
          return done(err);
        });
      });

      it('should be able to remember models', function (done) {
        var m = null;

        driver.createOttomanModel('RememberMe', {
          foo: 'string'
        }).then(function (model) {
          m = model;

          return driver.getModel('RememberMe');
        }).then(function (rememberedModel) {
          expect(rememberedModel).to.equal(m);
          done();
        }).catch(function (err) {
          return done(err);
        });
      });
    });

    describe('Indexes', function () {
      it('should add and remove an index', function (done) {
        var m = null;

        var indexExists = function indexExists(name) {
          return driver.getIndexes().then(function (results) {
            var found = false;
            console.log(JSON.stringify(results));
            console.log('Indexes: ' + results.rows.map(function (i) {
              return i.indexes.name;
            }).join(', '));

            for (var i = 0; i < results.rows.length; i++) {
              var z = results.rows[i];
              if (z.indexes.name === name) {
                found = true;
              }
            }

            return found;
          });
        };

        driver.createTable('Test', {
          name: 'string',
          favoriteAnimal: 'string'
        }).then(function (model) {
          m = model;
          return driver.addIndex('Test', 'Test_favoriteAnimalsIdx', 'favoriteAnimal');
        }).then(function () {
          return indexExists('Test_favoriteAnimalsIdx');
        }).then(function (exists) {
          expect(exists).to.be.true;
          return driver.removeIndex('Test_favoriteAnimalsIdx');
        }).then(function () {
          return indexExists('Test_favoriteAnimalsIdx');
        }).then(function (exists) {
          expect(exists).to.be.false;
          done();
        }).catch(function (err) {
          return done(err);
        });
      });
    });

    describe('Types', function () {
      var m = null;

      before(function (done) {
        var model = driver.createTable('TypeTest', {
          changeMeToNum: 'string',
          changeMeToStr: 'boolean',
          changeMeToBool: 'number',
          changeStrToBool: 'string'
        }).then(function (model) {
          m = model;

          var inst = new model({
            changeMeToNum: '500',
            changeMeToStr: true,
            changeMeToBool: 0,
            changeStrToBool: ''
          });

          return new Promise(function (resolve, reject) {
            return inst.save(function (err) {
              if (err) {
                return reject(err);
              }
              return resolve(inst);
            });
          });
        }).then(function () {
          done();
        });
      });

      it('shouldn\'t convert bad types', function (done) {
        driver.changeType('TypeTest', 'changeMeToNum', 'blark', 'blork').then(function (r) {
          return done('Converted bad type');
        }).catch(function () {
          return done();
        });
      });

      it('shouldn\'t convert unsupported types', function (done) {
        driver.changeType('TypeTest', 'changeMeToNum', 'Date', 'boolean').then(function (r) {
          return done('Converted bad type pair');
        }).catch(function () {
          return done();
        });
      });

      var changeTest = function changeTest(field, fromT, toT, done) {
        driver.changeType('TypeTest', field, fromT, toT).then(function () {
          return driver.runN1ql('SELECT TYPE(' + field + ') AS t\n              FROM `' + config.migrationBucket + '`\n              WHERE _type=\'TypeTest\'');
        }).then(function (r) {
          console.log(JSON.stringify(r));
          expect(r.rows.length).to.be.above(0);

          r.rows.forEach(function (item) {
            expect(item.t).to.equal(toT);
          });

          done();
        }).catch(function (err) {
          console.error('Failed type test: ' + err);
          done(err);
        });
      };

      it('should change strings to booleans', function (done) {
        changeTest('changeStrToBool', 'string', 'boolean', done);
      });

      it('should change strings to numbers', function (done) {
        changeTest('changeMeToNum', 'string', 'number', done);
      });

      it('should change booleans to strings', function (done) {
        changeTest('changeMeToStr', 'boolean', 'string', done);
      });

      it('should change numbers to booleans', function (done) {
        changeTest('changeMeToBool', 'number', 'boolean', done);
      });
    });

    describe('Buckets', function () {
      it('should be able to create and remove a bucket', function (done) {
        driver.createBucket('tmpTest').then(function (bucket) {
          expect(bucket).to.be.an('object');
          expect(bucket._name).to.equal('tmpTest');

          return driver.getBucketNames();
        }).then(function (buckets) {
          var bucketNames = buckets.map(function (i) {
            return i.name;
          });
          console.log(JSON.stringify(bucketNames));
          expect(bucketNames.indexOf('tmpTest')).to.be.above(-1);
          return driver.dropBucket('tmpTest');
        }).then(function (didDrop) {
          expect(didDrop).to.be.true;
          return driver.getBucketNames();
        }).then(function (buckets) {
          var bucketNames = buckets.map(function (i) {
            return i.name;
          });
          expect(bucketNames.indexOf('tmpTest')).to.be.below(0);
          done();
        }).catch(function (err) {
          return done(err);
        });
      });

      it('should expose working withBucket', function (done) {
        driver.withBucket('default').then(function () {
          return done();
        }).catch(function (err) {
          return done(err);
        });
      });

      it('should reject promise on bad bucket', function (done) {
        driver.withBucket('blizzlenert-non-existant').then(function () {
          return done(new Error('WTF, that bucket doesn\'t exist'));
        }).catch(function () {
          return done();
        });
      });

      it('should be able to list buckets', function (done) {
        driver.getBucketNames().then(function (buckets) {
          expect(buckets).to.be.an('array');
          expect(buckets.length).to.be.above(0);
          done();
        }).catch(function (err) {
          return done(err);
        });
      });
    });

    describe('Compatibility', function () {
      it('all is a synonym for runN1ql', function (done) {
        driver.all('SELECT * FROM system:indexes').then(function (results) {
          expect(results.rows).to.be.an('array');
          expect(results.meta).to.be.an('object');
          done();
        }).catch(function (err) {
          return done(err);
        });
      });

      it('runSql is a synonym for runN1ql', function (done) {
        driver.runSql('SELECT * FROM system:indexes').then(function (results) {
          expect(results.rows).to.be.an('array');
          expect(results.meta).to.be.an('object');
          done();
        }).catch(function (err) {
          return done(err);
        });
      });

      it('fails on bad n1ql', function (done) {
        driver.runN1ql('Blizzle-nert, this is not good n1ql').then(function (result) {
          return done('WTF: how did it run that??? ' + result);
        }).catch(function (err) {
          return done();
        });
      });
    });

    describe('Paths', function () {
      var FooMaticSchema = {
        name: 'string',
        foo: 'boolean',
        subDoc: {
          x: 'string'
        }
      };
      var FooMaticInstance = {
        name: 'John',
        foo: true,
        subDoc: {
          x: 'Hi'
        }
      };

      var scenarios = [{ name: 'simple path', attr: 'likesPie', modelName: 'FooMatic' }, { name: 'compound path', attr: 'subDoc.somethingElse', modelName: 'FooMatic2' }];

      scenarios.forEach(function (scenario) {
        it('should add and remove an ottoman ' + scenario.name, function (done) {
          var m = null;

          driver.createTable(scenario.modelName, FooMaticSchema).then(function (model) {
            m = model;

            var inst = new m(FooMaticInstance);

            return new Promise(function (resolve, reject) {
              return inst.save(function (err) {
                if (err) {
                  return reject(err);
                }
                return resolve(inst);
              });
            });
          }).then(function () {
            return driver.addOttomanPath(scenario.modelName, scenario.attr);
          }).then(function () {
            return driver.runN1ql('SELECT count(*) as x FROM `' + driver.activeBucketName() + '`\n                                    WHERE _type=\'' + scenario.modelName + '\' and ' + scenario.attr + ' IS NOT MISSING');
          }).then(function (results) {
            console.log(JSON.stringify(results));
            expect(results.rows[0].x).to.be.above(0);
            return driver.removeOttomanPath(scenario.modelName, scenario.attr);
          }).then(function () {
            return driver.runN1ql('SELECT count(*) as x FROM `' + config.migrationBucket + '`\n                                    WHERE _type=\'' + scenario.modelName + '\' and ' + scenario.attr + ' IS NOT MISSING');
          }).then(function (results) {
            console.log(JSON.stringify(results));
            expect(results.rows[0].x).to.equal(0);
            done();
          }).catch(function (err) {
            return done(err);
          });
        });
      });

      it('should rename an ottoman path', function (done) {
        var m = null;

        driver.createTable('RenamePathTest', {
          name: 'string',
          oldName: 'string'
        }).then(function (model) {
          m = model;
          var inst = new m({ name: 'John', oldName: 'something' });

          return new Promise(function (resolve, reject) {
            return inst.save(function (err) {
              if (err) {
                return reject(err);
              }
              return resolve(inst);
            });
          });
        }).then(function () {
          return driver.renameOttomanPath('RenamePathTest', 'oldName', 'newName');
        }).then(function () {
          return driver.runN1ql('SELECT count(*) as x FROM ' + config.migrationBucket + '\n                                    WHERE _type=\'RenamePathTest\' AND oldName IS MISSING AND\n                                    newName IS NOT MISSING');
        }).then(function (results) {
          expect(results.rows[0].x).to.be.above(0);
          done();
        }).catch(function (err) {
          return done(err);
        });
      });
    });

    after(function (done) {
      console.log('AFTER DRIVER: ' + driver);
      driver.deleteMigration('TestMigration').then(driver.allLoadedMigrations).then(function (migrations) {
        // After we've deleted it, it shouldn't be there.
        for (var i = 0; i < migrations.length; i++) {
          expect(migrations[i].name).to.not.equal('TestMigration');
        }

        done();
      }).catch(function (zerr) {
        return done(zerr);
      });
    });
  });
});
//# sourceMappingURL=../maps/test/basic_spec.js.map
